package ${basePackage}.frame.excel;

import com.alibaba.excel.EasyExcel;
import com.alibaba.excel.metadata.Head;
import com.alibaba.excel.support.ExcelTypeEnum;
import com.alibaba.excel.write.style.column.LongestMatchColumnWidthStyleStrategy;
import org.apache.poi.ss.usermodel.CellStyle;
import ${basePackage}.frame.excel.annotation.ExcelNote;
import ${basePackage}.frame.excel.annotation.ExcelSelect;
import ${basePackage}.frame.excel.annotation.ExcelSheet;
import ${basePackage}.frame.excel.converter.BooleanConverter;
import ${basePackage}.frame.excel.exception.TemplateNotMatchException;
import ${basePackage}.frame.excel.handler.HeadWriteHandler;
import ${basePackage}.frame.excel.handler.WCellWriteHandler;
import ${basePackage}.frame.excel.handler.WRowWriteHandler;
import ${basePackage}.frame.excel.handler.WSheetWriteHandler;
import ${basePackage}.frame.excel.listener.CheckHeadReadListener;
import ${basePackage}.frame.utils.ClassUtil;
import ${basePackage}.frame.utils.FileUtil;
import ${basePackage}.frame.utils.ValidationUtil;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class WExcel<T> {

    /**
     * 工作簿名称,可通过注解添加{@link ExcelSelect}.
     */
    private String name;

    /**
     * 模板类
     */
    private Class<T> clazz;

    /**
     * 表头的集合{@link com.alibaba.excel.annotation.ExcelProperty}.
     */
    private List<WHead> wHeads;

    /**
     * 数据
     */
    private List<T> data;

    /**
     * 错误Map key:行号,String:错误集
     */
    private Map<Integer, String> errMap;

    /**
     * 样式合集,请参考{@link WSheetWriteHandler#beforeSheetCreate}.
     * <p>
     * 说明:styles[0]为表头样式，styles[1]错误单元样式
     */
    private CellStyle[] styles = new CellStyle[2];

    /**
     * 根据模板类初始化
     */
    public WExcel(Class<T> clazz) {
        this.clazz = clazz;
        this.wHeads = new ArrayList<>();
        this.data = new ArrayList<>();
        this.errMap = new LinkedHashMap<>();
        //获取工作簿名称，没有则以类名为默认工作簿名称
        if (clazz.isAnnotationPresent(ExcelSheet.class)) {
            ExcelSheet ExcelSheet = clazz.getAnnotation(ExcelSheet.class);
            this.name = ExcelSheet.value();
        } else {
            this.name = clazz.getSimpleName();
        }

        // 获取实体所有字段
        Field[] fields = ClassUtil.getFields(clazz, true);
        Map<String, Field> fieldMap = new HashMap<>();
        for (Field field : fields) {
            fieldMap.put(field.getName(), field);
        }


        Map<String, WHead> wHeadMap = new LinkedHashMap<>();
        // 获取导出Head
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        EasyExcel.write(outputStream)
                .head(this.clazz)
                .useDefaultStyle(false)
                .registerConverter(new BooleanConverter())
                .registerWriteHandler(new HeadWriteHandler() {
                    @Override
                    protected void handlerHead(Head head) {


                        WHead wHead = new WHead(head);
                        Field field = fieldMap.get(wHead.getHead().getFieldName());

                        // 获取HeadName
                        wHead.setName(head.getHeadNameList().get(head.getHeadNameList().size() - 1));
                        // 获取注释
                        if (field.isAnnotationPresent(ExcelNote.class)) {
                            wHead.setNote(field.getAnnotation(ExcelNote.class).value());
                        }
                        // 获取下拉列表
                        if (field.isAnnotationPresent(ExcelSelect.class)) {
                            wHead.setSelectList(field.getAnnotation(ExcelSelect.class).value());
                        }

                        wHeadMap.put(head.getFieldName(), wHead);
                    }
                })
                .sheet(this.name).doWrite(Collections.EMPTY_LIST);
        wHeads.addAll(wHeadMap.values());
    }

    /**
     * 添加数据集
     *
     * @param os        待添加数据集
     * @param processor 执行器
     * @return WExcel
     */
    public WExcel addDatas(List<T> os, Processor<T> processor) {
        for (T o : os) {
            addData(o, processor);
        }
        return this;
    }

    /**
     * 添加数据
     *
     * @param o         待添加数据
     * @param processor 执行器
     * @return WExcel
     */
    public WExcel addData(T o, Processor<T> processor) {
        // 注解检查验证
        List<String> validate = ValidationUtil.validate(o);
        // 处理器执行
        if ((validate == null || validate.isEmpty()) && processor != null) {
            List<String> exec = processor.exec(o, this.data.size());
            if (exec != null && exec.size() > 0) {
                errMap.put(data.size() - 1, String.join("\r\n", exec));
            }
        }
        this.data.add(o);
        return this;
    }

    /**
     * @return 所有已添加数据
     */
    public List<T> getDatas() {
        return this.data;
    }

    /**
     * @return Excel文件字节数组
     */
    public byte[] getByte() {
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        EasyExcel.write(outputStream)
                .head(this.clazz)
                .useDefaultStyle(false)
                .registerConverter(new BooleanConverter())
                .registerWriteHandler(new LongestMatchColumnWidthStyleStrategy())
                .registerWriteHandler(new WSheetWriteHandler(wHeads, styles))
                .registerWriteHandler(new WRowWriteHandler(wHeads.size(), errMap))
                .registerWriteHandler(new WCellWriteHandler(styles, errMap))
                .sheet(this.name).doWrite(data);
        return outputStream.toByteArray();
    }

    /**
     * 将Excel文件输出到文件
     *
     * @param file 文件
     */
    public void toFile(File file) {
        byte[] aByte = getByte();
        FileUtil.writeBytesToFile(aByte, file);
    }

    /**
     * @param file      文件
     * @param processor 执行器
     * @return WExcel
     * @throws TemplateNotMatchException
     * @throws IOException
     */
    public WExcel<T> read(File file, Processor<T> processor) throws TemplateNotMatchException, IOException {
        ExcelTypeEnum et;
        if (file.getName().endsWith(".xls")) {
            et = ExcelTypeEnum.XLS;
        } else if (file.getName().endsWith(".xlsx")) {
            et = ExcelTypeEnum.XLSX;
        } else {
            throw new RuntimeException("The file is not an excel file!");
        }
        byte[] bytes = FileUtil.readFileToByteArray(file);
        return read(bytes, et, processor);
    }

    /**
     * @param bytes     字节数组
     * @param processor 执行器
     * @return WExcel
     * @throws TemplateNotMatchException
     * @throws IOException
     */
    public WExcel<T> read(byte[] bytes, ExcelTypeEnum et, Processor<T> processor) throws TemplateNotMatchException, IOException {
        ByteArrayInputStream stream = new ByteArrayInputStream(bytes);
        return read(stream, et, processor);
    }

    /**
     * @param in        输入流
     * @param processor 执行器
     * @return WExcel
     * @throws TemplateNotMatchException
     * @throws IOException
     */
    public WExcel<T> read(InputStream in, ExcelTypeEnum et, Processor<T> processor) throws TemplateNotMatchException {
        final TemplateNotMatchException[] e = new TemplateNotMatchException[1];
        // 读取Head 验证模板是否符合情况
        EasyExcel.read(in)
                .excelType(et)
                .head(this.clazz)
                .registerConverter(new BooleanConverter())
                .registerReadListener(new CheckHeadReadListener(wHeads, e))
//                .registerReadListener(new AnalysisEventListener<T>() {
//
//                    @Override
//                    public void invokeHeadMap(Map<Integer, String> headMap, AnalysisContext context) {
//
//                    }
//
//                    @Override
//                    public void invoke(T o, AnalysisContext analysisContext) {
//                        if (processor != null) {
//                            List<String> exec = processor.exec(o, data.size());
//                            if (exec != null && exec.size() > 0) {
//                                errMap.put(data.size(), String.join(",", exec));
//                            }
//                        }
//                        data.add(o);
//                    }
//
//                    @Override
//                    public void doAfterAllAnalysed(AnalysisContext analysisContext) {
//
//                    }
//
//                    @Override
//                    public boolean hasNext(AnalysisContext context) {
//                        return super.hasNext(context);
//                    }
//                })
                .sheet().doRead();
        if (e[0] != null) {
            throw e[0];
        }
        return this;
    }

    public interface Processor<T> {
        List<String> exec(T t, int index);
    }

    public static void main(String[] args) throws TemplateNotMatchException, IOException {
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        System.out.println("数据创建开始:" + dateFormat.format(new Date()));
        // 写
        List<DemoData> visitors = new ArrayList<>();
        for (int i = 0; i < 100; i++) {
            DemoData demoData = new DemoData();
            demoData.setStr1("我是标题Str1" + i);
            demoData.setStr2("我是标题Str2" + i);
            demoData.setStr3("我是标题Str3" + i);
            demoData.setStr4("我是标题Str4" + i);
            demoData.setSint((short) 1);
            demoData.setMint(1);
            demoData.setLint(1L);
            demoData.setDbn(0.01);
            demoData.setFln(0.01f);
            demoData.setDate(new Date());
            visitors.add(demoData);
        }
        System.out.println("数据创建完成:" + dateFormat.format(new Date()));
//        new WExcel<>(DemoData.class).addDatas(visitors, null).toFile(new File("D:\\xx.xlsx"));
        System.out.println("数据导出完成:" + dateFormat.format(new Date()));

        System.out.println();

        // 读
        System.out.println("数据导入开始:" + dateFormat.format(new Date()));
        WExcel<DemoData> wExcel = new WExcel<>(DemoData.class).read(new File("D:\\xx.xlsx"), new Processor<DemoData>() {
            @Override
            public List<String> exec(DemoData demoData, int index) {
                if (index % 50 == 0) {
//                    System.out.println(MapperUtil.toJson(demoData));
                    return Arrays.asList("测试,第" + index + "条有错误");
                }
                return null;
            }
        });
        System.out.println("数据导入完成:" + dateFormat.format(new Date()));
        System.out.println();
        System.out.println("错误导出开始:" + dateFormat.format(new Date()));
        wExcel.toFile(new File("D:\\xx-err.xlsx"));
        System.out.println("错误导出结束:" + dateFormat.format(new Date()));
        System.out.println();
    }
}
